@@ -231,6 +231,25 @@ impl From<std::path::StripPrefixError> for FileManifestError {
     }
 }
 
+/// Normalize a path or error on validation failure.
+///
+/// This is called before inserting paths into a [FileManifest].
+pub fn normalize_path(path: &Path) -> Result<PathBuf, FileManifestError> {
+    // Always store UNIX style directory separators.
+    let path_s = format!("{}", path.display()).replace('\\', "/");
+
+    if path_s.contains("..") {
+        return Err(FileManifestError::IllegalRelativePath(path_s));
+    }
+
+    // is_absolute() on Windows doesn't check for leading /.
+    if path_s.starts_with('/') || path.is_absolute() {
+        return Err(FileManifestError::IllegalAbsolutePath(path_s));
+    }
+
+    Ok(PathBuf::from(path_s))
+}
+
 /// Represents a collection of files.
 ///
 /// Files are keyed by their path. The file content is abstract and can be
@@ -264,7 +283,7 @@ impl FileManifest {
         let add_path = path.strip_prefix(strip_prefix)?;
 
         self.files
-            .insert(add_path.to_path_buf(), FileEntry::try_from(path)?);
+            .insert(normalize_path(add_path)?, FileEntry::try_from(path)?);
 
         Ok(())
     }
@@ -285,7 +304,7 @@ impl FileManifest {
         let add_path = path.strip_prefix(strip_prefix)?;
 
         let entry = FileEntry::try_from(path)?.to_memory()?;
-        self.files.insert(add_path.to_path_buf(), entry);
+        self.files.insert(normalize_path(add_path)?, entry);
 
         Ok(())
     }
@@ -298,19 +317,8 @@ impl FileManifest {
         path: impl AsRef<Path>,
         entry: impl Into<FileEntry>,
     ) -> Result<(), FileManifestError> {
-        let path = path.as_ref();
-        let path_s = path.display().to_string();
-
-        if path_s.contains("..") {
-            return Err(FileManifestError::IllegalRelativePath(path_s));
-        }
-
-        // is_absolute() on Windows doesn't check for leading /.
-        if path_s.starts_with('/') || path.is_absolute() {
-            return Err(FileManifestError::IllegalAbsolutePath(path_s));
-        }
-
-        self.files.insert(path.to_path_buf(), entry.into());
+        self.files
+            .insert(normalize_path(path.as_ref())?, entry.into());
 
         Ok(())
     }
@@ -395,13 +403,20 @@ impl FileManifest {
         self.files.iter().map(|(k, v)| File::new(k, v.clone()))
     }
 
+    /// Remove an entry from this manifest.
+    pub fn remove(&mut self, path: impl AsRef<Path>) -> Option<FileEntry> {
+        self.files.remove(path.as_ref())
+    }
+
     /// Obtain entries in this manifest grouped by directory.
     ///
     /// The returned map has keys corresponding to the relative directory and
     /// values of files in that directory.
     ///
     /// The root directory is modeled by the `None` key.
-    pub fn entries_by_directory(&self) -> BTreeMap<Option<&Path>, BTreeMap<&OsStr, &FileEntry>> {
+    pub fn entries_by_directory(
+        &self,
+    ) -> BTreeMap<Option<&Path>, BTreeMap<&OsStr, (&Path, &FileEntry)>> {
         let mut res = BTreeMap::new();
 
         for (path, content) in &self.files {
@@ -418,7 +433,7 @@ impl FileManifest {
             let filename = path.file_name().unwrap();
 
             let entry = res.entry(parent).or_insert_with(BTreeMap::new);
-            entry.insert(filename, content);
+            entry.insert(filename, (path.as_path(), content));
 
             // Ensure there are keys for all parents.
             if let Some(parent) = parent {
@@ -626,20 +641,41 @@ mod tests {
 
         assert_eq!(
             entries.get(&None).unwrap(),
-            &[(OsStr::new("root.txt"), &c),].iter().cloned().collect()
+            &[(
+                OsStr::new("root.txt"),
+                (PathBuf::from("root.txt").as_path(), &c)
+            ),]
+            .iter()
+            .cloned()
+            .collect()
         );
         assert_eq!(
             entries.get(&Some(Path::new("dir0"))).unwrap(),
-            &[(OsStr::new("dir0_file0.txt"), &c)]
-                .iter()
-                .cloned()
-                .collect()
+            &[(
+                OsStr::new("dir0_file0.txt"),
+                (PathBuf::from("dir0/dir0_file0.txt").as_path(), &c)
+            )]
+            .iter()
+            .cloned()
+            .collect()
         );
         assert_eq!(
             entries.get(&Some(Path::new("dir0/child0"))).unwrap(),
             &[
-                (OsStr::new("dir0_child0_file0.txt"), &c),
-                (OsStr::new("dir0_child0_file1.txt"), &c)
+                (
+                    OsStr::new("dir0_child0_file0.txt"),
+                    (
+                        PathBuf::from("dir0/child0/dir0_child0_file0.txt").as_path(),
+                        &c
+                    )
+                ),
+                (
+                    OsStr::new("dir0_child0_file1.txt"),
+                    (
+                        PathBuf::from("dir0/child0/dir0_child0_file1.txt").as_path(),
+                        &c
+                    )
+                )
             ]
             .iter()
             .cloned()
@@ -647,17 +683,131 @@ mod tests {
         );
         assert_eq!(
             entries.get(&Some(Path::new("dir0/child1"))).unwrap(),
-            &[(OsStr::new("dir0_child1_file0.txt"), &c)]
-                .iter()
-                .cloned()
-                .collect()
+            &[(
+                OsStr::new("dir0_child1_file0.txt"),
+                (
+                    PathBuf::from("dir0/child1/dir0_child1_file0.txt").as_path(),
+                    &c
+                )
+            )]
+            .iter()
+            .cloned()
+            .collect()
         );
         assert_eq!(
             entries.get(&Some(Path::new("dir1/child0"))).unwrap(),
-            &[(OsStr::new("dir1_child0_file0.txt"), &c)]
-                .iter()
-                .cloned()
-                .collect()
+            &[(
+                OsStr::new("dir1_child0_file0.txt"),
+                (
+                    PathBuf::from("dir1/child0/dir1_child0_file0.txt").as_path(),
+                    &c
+                )
+            )]
+            .iter()
+            .cloned()
+            .collect()
+        );
+
+        Ok(())
+    }
+
+    #[test]
+    fn test_entries_by_directory_windows() -> Result<(), FileManifestError> {
+        let c = FileEntry {
+            data: vec![42].into(),
+            executable: false,
+        };
+
+        let mut m = FileManifest::default();
+        m.add_file_entry(Path::new("root.txt"), c.clone())?;
+        m.add_file_entry(Path::new("dir0\\dir0_file0.txt"), c.clone())?;
+        m.add_file_entry(Path::new("dir0\\child0\\dir0_child0_file0.txt"), c.clone())?;
+        m.add_file_entry(Path::new("dir0\\child0\\dir0_child0_file1.txt"), c.clone())?;
+        m.add_file_entry(Path::new("dir0\\child1\\dir0_child1_file0.txt"), c.clone())?;
+        m.add_file_entry(Path::new("dir1\\child0\\dir1_child0_file0.txt"), c.clone())?;
+
+        let entries = m.entries_by_directory();
+
+        assert_eq!(entries.keys().count(), 6);
+        assert_eq!(
+            entries.keys().collect::<Vec<_>>(),
+            vec![
+                &None,
+                &Some(Path::new("dir0")),
+                &Some(Path::new("dir0/child0")),
+                &Some(Path::new("dir0/child1")),
+                &Some(Path::new("dir1")),
+                &Some(Path::new("dir1/child0")),
+            ]
+        );
+
+        assert_eq!(
+            entries.get(&None).unwrap(),
+            &[(
+                OsStr::new("root.txt"),
+                (PathBuf::from("root.txt").as_path(), &c)
+            ),]
+            .iter()
+            .cloned()
+            .collect()
+        );
+        assert_eq!(
+            entries.get(&Some(Path::new("dir0"))).unwrap(),
+            &[(
+                OsStr::new("dir0_file0.txt"),
+                (PathBuf::from("dir0/dir0_file0.txt").as_path(), &c)
+            )]
+            .iter()
+            .cloned()
+            .collect()
+        );
+        assert_eq!(
+            entries.get(&Some(Path::new("dir0/child0"))).unwrap(),
+            &[
+                (
+                    OsStr::new("dir0_child0_file0.txt"),
+                    (
+                        PathBuf::from("dir0/child0/dir0_child0_file0.txt").as_path(),
+                        &c
+                    )
+                ),
+                (
+                    OsStr::new("dir0_child0_file1.txt"),
+                    (
+                        PathBuf::from("dir0/child0/dir0_child0_file1.txt").as_path(),
+                        &c
+                    )
+                )
+            ]
+            .iter()
+            .cloned()
+            .collect()
+        );
+        assert_eq!(
+            entries.get(&Some(Path::new("dir0/child1"))).unwrap(),
+            &[(
+                OsStr::new("dir0_child1_file0.txt"),
+                (
+                    PathBuf::from("dir0/child1/dir0_child1_file0.txt").as_path(),
+                    &c
+                )
+            )]
+            .iter()
+            .cloned()
+            .collect()
+        );
+        assert_eq!(
+            entries.get(&Some(Path::new("dir1/child0"))).unwrap(),
+            &[(
+                OsStr::new("dir1_child0_file0.txt"),
+                (
+                    PathBuf::from("dir1/child0/dir1_child0_file0.txt").as_path(),
+                    &c
+                )
+            )]
+            .iter()
+            .cloned()
+            .collect()
         );
 
         Ok(())
